package src.hw6S09;

public class DuctTree {
	DuctNode root;						// Package access for all data members
	double cd;							// Cost ($) per sq meter of duct
	double crf;							// Capital recovery factor (amortization, annual)
	double ce;							// Cost ($) per kwh of electricity to power fan
	double[] solutionPressures;			// Output of algorithm is list of pressures for each duct. From pressure, we can compute diameter, velocity, cost
	int nbrNodes;						// Nodes in tree. Assume nodes numbered consecutively, and node 0 is unused

	public DuctTree(double cd, double crf, double ce) {
		root= null;
		this.cd= cd;
		this.crf= crf;
		this.ce= ce;					// Allocate memory for solutions array after number of nodes is known
	}

	public void preorder() {
		if (root != null)
			root.traversePreorder();
	}

	public void postorder() {
		if (root != null)
			root.traversePostorder();
	}

	public double processRootNode() {					
		solutionPressures= new double[nbrNodes+1];		// Allocate memory for solution, to be written in backward pass by preorder traverse, to be done next
		double minRootCost= DuctNode.INFINITY;
		int minRootCostPressure= 0;
		// Operating costs: flow q * electric cost ce * hours per year/ efficiency. Multiply by .001 to convert from watt-hr to kwh. This multiples pressure.
		double opCostFactor= 0.001 * root.q * ce * 8640.0/0.64;

		// Find lowest total capital plus operating cost
		for (int i= 1; i < DuctNode.MAX_PRESSURE/DuctNode.PRESSURE_INCR; i++) {
			if (root.cost[i] < DuctNode.INFINITY) {
				double capitalCost= root.cost[i]*crf;
				double opCost= opCostFactor*i*DuctNode.PRESSURE_INCR;
				double totalCost= capitalCost + opCost;	

				if (totalCost < minRootCost) {
					minRootCost= totalCost;
					minRootCostPressure= i* DuctNode.PRESSURE_INCR;
				}
			}
		}
		root.minCostPressure= minRootCostPressure;
		System.out.println(" Op cost: " + opCostFactor*minRootCostPressure);
		System.out.println(" Capital cost: " + (minRootCost - opCostFactor*minRootCostPressure));
		return minRootCost;
	}
	
	public class DuctNode {					// Data members are package access
		int ductNbr;						// Duct number
		double q;							// Air flow (m^3/sec)
		double length;						// Duct length (m)
		double[] cost;						// Cost so far at each pressure at this node
		int[] predecessor;					// Pressure in downstream node that made up this pressure
		
		DuctNode left;
		DuctNode right;
		double minCostPressure;				// Pressure at which minimum cost occurs (output)

		double minD;						// Min duct diameter (m)
		double maxD;						// Max duct diameter (m)
		double minV;						// Min velocity in duct (m/sec)
		double maxV;						// Max velocity in duct (m/sec)

		public static final double TOLERANCE= 1E-15;
		public static final int MAX_PRESSURE= 150;
		public static final int PRESSURE_INCR= 5;
		public static final double INFINITY= Double.MAX_VALUE;

		public DuctNode(int dn, DuctNode left, DuctNode right, double q, double length, double minD, double maxD, double minV, double maxV) {
			ductNbr= dn;
			this.left= left;
			this.right= right;
			this.q= q;
			this.length= length;
			this.minD= minD;
			this.maxD= maxD;
			this.minV= minV;
			this.maxV= maxV;
			cost= new double[MAX_PRESSURE/PRESSURE_INCR];
			for (int i= 0; i < MAX_PRESSURE/PRESSURE_INCR; i++)
				cost[i]= INFINITY;
			predecessor= new int[MAX_PRESSURE/PRESSURE_INCR];
			for (int i= 0; i < MAX_PRESSURE/PRESSURE_INCR; i++)
				predecessor[i]= -1;
		}

		// Approximation for problem set. True equations are more complex, but tractable
		public double getDuctSize(double p) {
			return Math.pow(0.03*q*q*length/p, 0.2);
		}

		public double getCost(double d) {
			return Math.PI * d * length * cd;		// Dollars, per sq m of duct
		}

		public double getVelocity(double d) {
			return 4.0*q/(Math.PI*d*d);
		}

		public void traversePostorder() {
			nbrNodes++;										// Count number of nodes in tree
			if (left != null) left.traversePostorder();
			if (right != null) right.traversePostorder();

			// Add the flow from left and right child node. If node is not a leaf, its own flow is zero. If node is a leaf, q > 0 from the input
			if (left != null) q += left.q;
			if (right != null) q += right.q;

			// Compute the pressure and cost pairs for the duct going upstream (toward the fan) from this node
			double pressure= 0.0;
			double[] ownCost= new double[MAX_PRESSURE/PRESSURE_INCR];
			for (int i= 0; i < MAX_PRESSURE/PRESSURE_INCR; i++)
				ownCost[i]= INFINITY;
			for (int i= 1; i < MAX_PRESSURE/PRESSURE_INCR; i++) {			
				pressure = i * PRESSURE_INCR;
				double ductSize= getDuctSize(pressure);
				double dcost= getCost(ductSize);
				double velocity= getVelocity(ductSize);
				if (velocity + TOLERANCE >= minV  &&  velocity <= maxV + TOLERANCE && ductSize <= maxD && ductSize >= minD)
					ownCost[i]= dcost; 
			}
			// Compute the pressures of the flows coming the the child nodes. The pressures from left and right node must be equal to be feasible.
			// Both the right and left child must have the pressure as feasible (within duct size and velocity limits); otherwise it's ignored
			double[] childCost= new double[MAX_PRESSURE/PRESSURE_INCR];
			for (int i= 0; i < MAX_PRESSURE/PRESSURE_INCR; i++)
				childCost[i]= INFINITY;
			if (left != null && right != null) { 			// Merge current duct with left child duct
				for (int i= 1; i < MAX_PRESSURE/PRESSURE_INCR; i++) 
					if (left.cost[i] > 0 && right.cost[i] > 0)			// Both left and right must have this pressure as a feasible result
						childCost[i]= left.cost[i] + right.cost[i];

				// Now combine the child flows (already added together) with the upstream flow from the node. 
				// This is label correcting: find the lowest cost for the pressure that is the sum of the child pressure plus upstream pressure. When found, store predecessor
				for (int i= 1; i < MAX_PRESSURE/PRESSURE_INCR; i++) {			// Own cost--upstream duct
					for (int j= 1; j < MAX_PRESSURE/PRESSURE_INCR; j++) {		// Child cost-- ducts feeding into this node
						if ((i+j < MAX_PRESSURE/PRESSURE_INCR) && (ownCost[i] + childCost[j] < cost[i+j])) {
							cost[i+j]= ownCost[i] + childCost[j];
							predecessor[i+j]= j;
						}	
					}
				}
			} else {		// No child ducts. Just copy own cost into cost
				for (int i= 1; i < MAX_PRESSURE/PRESSURE_INCR; i++)
					cost[i]= ownCost[i];
			}
		}

		public void traversePreorder() {
			// Backward pass, read off solution
			// The parent has already set the minCostPressure
			// We subtract the child pressure to get this duct's pressure and then output its other characteristics
			int i= (int)minCostPressure/PRESSURE_INCR;			// Map between pressure and element number in predecessor array. Not elegant.
			int pred= predecessor[i];
			double predPressure= pred*PRESSURE_INCR;
			double thisDuctPressure= minCostPressure - predPressure;
			double d= getDuctSize(thisDuctPressure);
			double c= getCost(d);
			double v= getVelocity(d);
			solutionPressures[ductNbr]= thisDuctPressure;		// Store output. For convenience, we just output the results in the line below
			System.out.println(" Duct: " + ductNbr + " cost: " + c + " duct size: " + d + " velocity: " + v + " pressure chg: " + thisDuctPressure);
			// We next set the left and right child's minCostPressure
			if (left != null &&  right !=  null) {
				left.minCostPressure= predPressure;
				right.minCostPressure= predPressure;
			}
			if (left != null) left.traversePreorder();
			if (right != null) right.traversePreorder();
		}
	}
}